(* Daan Leijen 2001, daan@cs.uu.nl *)
open Sml

type 'a maybe = 
  | Just of 'a
  | Nothing


(* nesting of colors & sizes: <r>a</r><g>b</g><r>c</r> => <r>a<g>b</g>c</r> *)
let rec find_span pred x acc ts
  = match ts with
    | (t::tt) 
         -> (match t with
            | Tag (tags,body) when List.exists pred tags
                -> if (List.mem x tags)
                    then Just (List.rev acc,tags,body,tt)
                    else find_span pred x (t::acc) tt
            | other -> Nothing
            )
    | [] -> Nothing

let opt_nesting pred tags0 tt
  = if (not (List.exists pred tags0))
     then Nothing
     else match List.partition pred tags0 with
          | ([x],diff0) 
            -> (match find_span pred x [] tt with 
               | Just (xs,tags1,body1,ys)
                   -> let diff1 = List.filter (fun t -> t <> x) tags1 in
                      Just (x,diff0,xs,diff1,body1,ys)                   
               | other -> Nothing
               )
          | other -> Nothing

let find_longest_span s c tags0 tt
  = match (find_span is_size s [] tt, find_span is_color c [] tt) with
    | (Just (xs,_,_,_),Just (ys,_,_,_)) 
        -> if (List.length ys > List.length xs)
            then opt_nesting is_color tags0 tt
            else opt_nesting is_size  tags0 tt
    | (Just _,Nothing)
        -> opt_nesting is_size  tags0 tt
    | (Nothing,Just _)
        -> opt_nesting is_color tags0 tt
    | other
        -> Nothing

let opt_nesting_color_size tags0 tt
  = match (List.filter is_size tags0,List.filter is_color tags0) with
    | ([s],[c]) -> find_longest_span s c tags0 tt
    | ([s],[])  -> opt_nesting is_size tags0 tt
    | ([],[c])  -> opt_nesting is_color tags0 tt
    | other     -> Nothing

(* nesting of PL elements: <I><S>hi</S></I>foo<I><S>there</S></I> => <I><S>hi<PL>foo</PL>there</S></I> *)
let rec find_pl_span ctx pltags acc ts 
  = match ts with
    | (t::tt)
        -> (match t with
           | Tag (tags1,body1) 
              -> let union = List.filter (fun t -> List.mem t pltags) tags1 in
                 if (List.length union > 1 && next_ctxs ctx (union @ [PL]) = ctx)
                  then Just ([Tag ([PL],List.rev acc)],tags1,body1,tt)
                  else find_pl_span ctx pltags (t::acc) tt
           | other
              -> find_pl_span ctx pltags (t::acc) tt
           )
    | [] -> Nothing
                     
let opt_pl_nesting ctx tags0 tt
  = let pltags0 = List.filter is_pl_sensitive tags0 in
    if (List.length pltags0 > 1 && next_ctxs ctx (pltags0 @ [PL]) = ctx)
     then match find_pl_span ctx pltags0 [] tt with
          | Just (xs,tags1,body1,ys)
              -> let union = List.filter (fun t -> List.mem t tags1) pltags0 in
                 let diff0 = List.filter (fun t -> not (List.mem t union)) tags0 in
                 let diff1 = List.filter (fun t -> not (List.mem t union)) tags1 in
                 Just (union,diff0,xs,diff1,body1,ys)
          | other -> Nothing
     else Nothing
                 

(* optimize nestings: color, size and PL *)
let rec opt_nestings ctx ts
  = match ts with
    | []  -> ts
    | (t::tt) -> (match t with
                 | Tag (tags0,body0) 
                     -> ((match opt_nesting_color_size tags0 tt with
                        | Just (x,diff0,xs,diff1,body1,ys)
                            -> Tag ([x],opt_nestings ctx (Tag (diff0,body0)::xs @ [Tag (diff1,body1)]))
                                  ::opt_nestings ctx ys
                         
                               | Nothing
                                  -> (match opt_pl_nesting ctx tags0 tt with
                                     | Just (union,diff0,xs,diff1,body1,ys)
                                        -> Tag (union,opt_nestings ctx (Tag (diff0,body0)::xs @ [Tag (diff1,body1)]))
                                             ::opt_nestings ctx ys
                                     | Nothing
                                        -> t::opt_nestings ctx tt
                                     )
                               )
                        )
                 | other -> t::opt_nestings ctx tt
                 )

(* overlap inversion *)
let overlap_left ctx tags0 body0 tags1 body1
  = match body1 with
    | (Tag (tags10,body10)::tt1)
        -> let (union,diff0) = List.partition (fun t -> List.mem t tags10) tags0 in
           if (   size_of_tags union > size_of_tags tags1 
               && next_ctxs ctx (union @ tags1) = next_ctxs ctx (tags1 @ union) )
            then let diff10 = List.filter (fun t -> not (List.mem t union)) tags10 in
                 let t0a    = Tag (diff0,body0) in
                 let t0b    = Tag (tags1,[Tag (diff10,body10)]) in
                 Just (Tag (union,[t0a;t0b]),Tag (tags1,tt1))
            else Nothing
    | other
        -> Nothing

let overlap_right ctx tags0 body0 tags1 body1
  = if (null body0)
     then Nothing
     else match List.rev body0 with (* todo: use split_last *)
          | (Tag (tags2,body2)::tt0)
              -> let (union,diff1) = List.partition (fun t -> List.mem t tags2) tags1 in
                 if ( size_of_tags union > size_of_tags tags0 
                      && next_ctxs ctx (union @ tags0) = next_ctxs ctx (tags0 @ union))
                  then let diff2 = List.filter (fun t -> not (List.mem t union)) tags2 in
                       let t1b   = Tag (diff1,body1) in
                       let t1a   = Tag (tags0,[Tag (diff2,body2)]) in
                 Just (Tag (tags0,List.rev tt0), Tag (union,[t1a;t1b]))
            else Nothing
    | other
        -> Nothing
     

(* optimize neighbour elements *)
let rec opt_neighbour ctx t0 t1 tt
  = match (t0,t1) with
    | (Tag (tags0,body0),Tag (tags1,body1))
        -> (* try to collapse common tags *)
           let (union,diff0) = List.partition (fun t -> List.mem t tags1) tags0 in
           if (null union)
            then (* try overlap inversion *)
                 (match overlap_left ctx tags0 body0 tags1 body1 with
                 | Just (to0,to1) 
                      -> to0 :: opt_neighbours ctx (to1 :: tt)
                 | Nothing  
                      -> (match overlap_right ctx tags0 body0 tags1 body1 with
                         | Just (to0,to1)
                            -> to0 :: opt_neighbours ctx (to1 :: tt)
                         | Nothing 
                            -> t0::opt_neighbours ctx (t1::tt)
                         )
                 ) 
            else let diff1 = List.filter (fun t -> not (List.mem t tags0)) tags1 in
                 let t     = Tag (union,[Tag (diff0,body0);Tag (diff1,body1)]) in
                 opt_neighbours ctx (t::tt)
    | other
        -> t0::opt_neighbours ctx (t1::tt)

and opt_neighbours ctx ts
  = match ts with
    | (t0::t1::tt) -> opt_neighbour ctx t0 t1 tt
    | other        -> ts


(* flatten PL bindings: <I>a<PL>b</PL>c</I> => <I>a</I>b<I>c</I> *)
let rec opt_flatten_pl_body ctx tag tagacc acc ts
  = match ts with
    | (t::tt)
         -> (match t with
            | Tag (tags,body) when List.mem PL tags
                -> let diff = List.filter (fun t -> t <> PL) tags in
                   if (next_ctxs ctx (tag::tags) = next_ctxs ctx diff)
                    then opt_flatten_pl_body ctx tag [] (Tag (diff,body)::Tag ([tag],List.rev tagacc)::acc) tt
                    else opt_flatten_pl_body ctx tag (t::tagacc) acc  tt
            | other
                -> opt_flatten_pl_body ctx tag (t::tagacc) acc tt
            )
    | [] -> List.rev (Tag ([tag],List.rev tagacc) :: acc)

let has_pl_tag t
  = match t with
    | Tag (tags,body) -> List.mem PL tags
    | other           -> false

let opt_flatten_pl ctx tags body
  = match List.filter is_pl_sensitive tags with
    | [tag] when List.mem tag [I;B;S;U]
        -> let diff = List.filter (fun t -> t <> tag) tags in
           if (next_ctxs ctx (tags @ [PL]) = next_ctxs ctx diff && List.exists has_pl_tag body)
            then Tag (diff,opt_flatten_pl_body (next_ctxs ctx diff) tag [] [] body)
            else Tag (tags,body)
    | other
        -> Tag (tags,body)


(* "lift" optimization (EM/EM & PL), for example <EM><EM>a</EM>b</EM> => a<EM>b</EM>  *)
let opt_lift_em ctx tags1 tags0
  = if (List.mem EM tags1 && List.mem EM tags0)
     then let diff1 = List.filter (fun t -> t <> EM) tags1 in
          let diff0 = List.filter (fun t -> t <> EM) tags0 in
          if (next_ctxs ctx (tags1 @ tags0) = next_ctxs ctx (diff1 @ diff0))
           then Just (diff1,diff0)
           else Nothing
     else Nothing

let opt_lift_pl ctx tags1 tags0
  = if (List.mem PL tags0)
     then let (diff1,union1) = List.partition is_pl_sensitive tags1 in
          if (not (null union1) && next_ctxs ctx (tags1 @ tags0) = next_ctxs ctx (union1 @ tags0))
           then Just (union1,diff1)
           else Nothing
     else Nothing

let rec opt_lift is_rev ctx tags1 body1
  = let rev xs  = if (is_rev) then List.rev xs else xs in
    let next () = if (is_rev) then opt_flatten_pl ctx tags1 body1  (* in the end, try to flatten PL tags too *)
                              else opt_lift (not is_rev) ctx tags1 body1 in
    match rev body1 with
    | (Tag (tags0,body0)::tt)
        -> (match opt_lift_em ctx tags1 tags0 with
           | Just (diff1,diff0)
              -> Tag (diff1,rev [Tag (diff0,body0);Tag ([EM],rev tt)])
           | Nothing
              -> (match opt_lift_pl ctx tags1 tags0 with
                 | Just (union1,diff1)
                    -> Tag (union1, rev [Tag (tags0,body0);Tag (diff1,rev tt)])
                 | Nothing
                    -> next ()
                 )
           )
    | other -> next ()


(* optimize nodes *)
let rec opt_tree ctx t
  = match t with
    | Text txt 
        -> if (String.length txt = 0) then [] else  [t]
    | Tag (tags,[Text " "])
        -> (* redundant for space *)
           let tags' = List.filter (fun tag -> space_ctx (next_ctx ctx tag) <> space_ctx ctx) tags in
           if (null tags')
            then [Text " "]
            else [Tag (tags',[Text " "])]
    | Tag (tags,body)
        -> (* redundant: filter those tags that have no effect *)
           let tags' = List.filter (fun tag -> next_ctx ctx tag <> ctx) tags in
           assert (next_ctxs ctx tags' = next_ctxs ctx tags);
           if (null tags')
            then opt_body ctx body
            else (* try "lift" optimization *)
                 [opt_lift false ctx tags' (opt_body (next_ctxs ctx tags') body)]

(* optimize bodies
  - first optimize the elements of a body: opt_tree
  - then optimize neighbouring elements
  - then do nesting optimizations that can span multiple elements
*)
and opt_list ctx len0 ts0
  = let ts1  = opt_nestings ctx (opt_neighbours ctx ts0) in
    let len1 = List.length ts1 in
    if (len1 < len0)
     then opt_list ctx len1 ts1
     else ts1

and opt_body ctx ts
  = let ts1 = List.concat (List.map (opt_tree ctx) ts) in
    opt_list ctx (List.length ts1) ts1

(* perform local optimizations *)
let opt_local ts
  = opt_body root_ctx ts