(* Daan Leijen 2001, daan@cs.uu.nl *)
(* SML trees *)
type tag =
  | B
  | EM
  | I
  | PL
  | S
  | TT
  | U     
  | Size  of int
  | Color of char

type sml =
  | Tag  of tag list * sml list
  | Text of string

(* utility functions on trees *)
let string_of_tag tag
  = match tag with
    | B   -> "B"
    | EM  -> "EM"
    | I   -> "I"
    | PL  -> "PL"
    | S   -> "S"
    | TT  -> "TT"
    | U   -> "U"  
    | Size sz -> string_of_int sz
    | Color c -> String.make 1 c

let is_size tag
  = match tag with
    | Size sz -> true
    | other   -> false

let is_color tag
  = match tag with
    | Color c  -> true
    | other    -> false


let size_of_tag tag
  = let brackets = 5 in
    match tag with
    | EM  -> 4 + brackets
    | PL  -> 4 + brackets
    | TT  -> 4 + brackets
    | _   -> 2 + brackets

let size_of_tags tags
  = List.fold_left (fun n tag -> n + size_of_tag tag) 0 tags 

let rec size_of_trees ts
  = match ts with
    | []  -> 0
    | Text txt::tt  
          -> String.length txt + size_of_trees tt
    | Tag (tags,body)::tt
          -> size_of_tags tags + size_of_trees body + size_of_trees tt

let is_pl_sensitive tag
  = tag < Size 0


(* contexts *)
type context =
  { bold      : bool
  ; italic    : bool
  ;  teletype  : bool
  ; emph      : bool
  ; strong    : bool
  ; underline : int
  ; size      : int
  ; color     : char
  }


let root_ctx  
  = { bold=false; italic=false; teletype=false; emph=false
    ; strong=false; underline=0; size=10; color='?' }

let space_ctx ctx
  = { ctx with strong=false; emph=false; italic=false; bold=false
             ; color=if (ctx.underline=0) then 'w' else ctx.color }

let next_ctx ctx tag
  = match tag with
    | B       -> { ctx with bold=true }
    | I       -> { ctx with italic=true }
    | TT      -> { ctx with teletype=true }
    | S       -> { ctx with strong=true; emph=false }
    | EM      -> if (ctx.strong) then ctx else { ctx with emph = not ctx.emph }
    | PL      -> { ctx with bold=false; italic=false; teletype=false; emph=false
                          ; strong=false; underline=0 }
    | U       -> if (ctx.underline >= 3) then ctx else {ctx with underline=ctx.underline+1 }
    | Size sz -> { ctx with size = sz }
    | Color c -> { ctx with color = c }
 
let next_ctxs ctx tags
  = List.fold_left next_ctx ctx tags 


(* utility *)
let null xs
  = match xs with
    | []  -> true
    | _   -> false

let string_last s
  = assert (String.length s > 0);
    String.get s (String.length s -1)

let string_of_revlist_len len cs
  = let s   = String.create len in
    let rec create i xs
            = match xs with
              | []      -> assert (i= -1); s
              | (x::xt) -> String.set s i x; create (i-1) xt in
    create (len-1) cs



let string_of_list_len len cs
  = let s   = String.create len in
    let rec create i xs
            = match xs with
              | []      -> assert (i=len); s
              | (x::xt) -> String.set s i x; create (i+1) xt in
    create 0 cs

let string_of_list cs
  = string_of_list_len (List.length cs) cs


